/**	\file ComplexType.cpp
 **	\date  2008-05-26
 **	\author grymse@alhem.net
**/
/*
Copyright (C) 2008  Anders Hedstrom

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/
#include "ComplexType.h"
#include <Sockets/XmlNode.h>
#include "WsdlException.h"
#include "Element.h"
#include <Sockets/Debug.h>
#include "Split.h"
#include "Wsdl.h"
#include "Attribute.h"
#include "SimpleType.h"
#include "Application.h"


std::map<std::string, bool> ComplexType::m_b_generated;


ComplexType::ComplexType(const Wsdl& wsdl, const Xml::XmlNode& node, const std::string& name, const std::string& ns_prefix)
: TypeBase(wsdl, ns_prefix, Split(name).Type)
, m_node(node)
, m_name(!name.empty() ? name : node.PropertyExists("name") ? node.GetProperty("name") : "")
, m_b_is_sequence(false)
, m_b_is_all(false)
, m_b_complexContent(false)
, m_b_is_array(false)
{
	// check 'sequence'
	{
		Xml::XmlNode n(node, "sequence");
		if (n)
			m_b_is_sequence = true;
	}
	// check 'all'
	{
		Xml::XmlNode n(node, "all");
		if (n)
			m_b_is_all = true;
	}
	//
	{
		Xml::XmlNode n(node, "complexContent");
		if (n)
		{
			m_b_complexContent = true;
			Xml::XmlNode n(n, "restriction");
			if (n)
			{
				if (n.PropertyExists("base") && Split(n.GetProperty("base")).Type == "Array")
				{
					Xml::XmlNode n(n, "attribute");
					if (n && n.PropertyExists("arrayType"))
					{
						std::string tmp = n.GetProperty("arrayType");
						Parse pa(tmp, "[]");
						tmp = pa.getword();
						SetType( tmp );
						m_b_is_array = true;
					}
				}
			}
		}
	}
}


ComplexType::~ComplexType()
{
}


int ComplexType::NumberOfMembers() const
{
	Xml::XmlNode n(m_node, m_b_is_sequence ? "sequence" : "all");
	int q = 0;
	if (n)
	{
		Xml::XmlNode n(n, "element");
		while (n)
		{
			q++;
			++n;
		}
	}
	return q;
}


bool ComplexType::GetRef(std::string& ref) const
{
	Xml::XmlNode n(m_node, m_b_is_sequence ? "sequence" : "all");
	if (n)
	{
		Xml::XmlNode n(n, "element");
		while (n)
		{
			if (n.PropertyExists("ref"))
			{
				ref = n.GetProperty("ref");
				return true;
			}
			++n;
		}
	}
	return false;
}


std::list<std::pair<std::string, Xml::XmlNode> > ComplexType::GetMembers() const
{
	std::list<std::pair<std::string, Xml::XmlNode> > vec;
	Xml::XmlNode n(m_node, m_b_is_sequence ? "sequence" : "all");
	if (n)
	{
		Xml::XmlNode n(n, "element");
		while (n)
		{
			if (n.PropertyExists("ref"))
			{
				std::string name = n.GetProperty("ref");
				vec.push_back(std::pair<std::string, Xml::XmlNode>(name, n));
			}
			else
			{
				std::string name = n.GetProperty("name");
				vec.push_back(std::pair<std::string, Xml::XmlNode>(name, n));
			}
			++n;
		}
	}
	return vec;
}


std::list<std::pair<std::string, Xml::XmlNode> > ComplexType::GetAttributes() const
{
	std::list<std::pair<std::string, Xml::XmlNode> > vec;
	Xml::XmlNode n(m_node, "attribute");
	while (n)
	{
		std::string name = n.GetProperty("name");
		vec.push_back(std::pair<std::string, Xml::XmlNode>(name, n));
		++n;
	}
	return vec;
}


const std::string ComplexType::Documentation() const
{
	return TypeBase::Documentation(m_node);
}


const std::string ComplexType::CName() const
{
	return Split(m_name).CName1up;
}


bool ComplexType::OkToGenerate() const
{
	if (m_b_complexContent)
	{
		if (m_b_is_array)
		{
			try
			{
				Xml::XmlNode n = m_wsdl.FindComplexType( Type() );
				ComplexType t(m_wsdl, n, Split(Type()).Type, Split(Type()).NS);
				if (!t.IsGenerated())
				{
					return false;
				}
			}
			catch (const Exception& )
			{
			}
			return true;
		}
	}
	std::list<std::pair<std::string, Xml::XmlNode> > members = GetMembers();
	std::list<std::pair<std::string, Xml::XmlNode> > attrs = GetAttributes();
	for (std::list<std::pair<std::string, Xml::XmlNode> >::const_iterator it = members.begin(); it != members.end(); it++)
	{
		const std::pair<std::string, Xml::XmlNode>& ref = *it;
		Element el(m_wsdl, ref.second, "", NsPrefix());
		if (el.IsComplexType())
		{
			Xml::XmlNode n = m_wsdl.FindComplexType( el.Type() );
			ComplexType t(m_wsdl, n, Split(el.Type()).Type, Split(el.Type()).NS);
			std::string fqn = NsPrefix() + ":" + t.Name();
			std::string parent = m_wsdl.GetParent(fqn);
			if (!t.IsGenerated() && parent != Name())
			{
				return false;
			}
		}
		else
		{
		}
	}
	return true;
}


void ComplexType::SetGenerated(bool x)
{
	m_b_generated[m_name] = x; 
}


bool ComplexType::IsGenerated() const
{
	std::map<std::string, bool>::const_iterator it = m_b_generated.find(m_name);

	if (it != m_b_generated.end())
	{
		return true;
	}
	return false;
}


void ComplexType::CreateInterface(FILE *fil, int lvl) const
{
	std::string prefix;
	for (int i = 0; i < lvl; i++)
		prefix += "\t";
	if (m_b_complexContent)
	{
		if (m_b_is_array)
		{
			fprintf(fil, "%stypedef %s<%s> %s;\n", TypeBase::CType().c_str(), m_wsdl.GetApplication().ListType().c_str(), CName().c_str());
		}
	}

	std::list<std::pair<std::string, Xml::XmlNode> > members = GetMembers();
	std::list<std::pair<std::string, Xml::XmlNode> > attrs = GetAttributes();
	if (members.empty() && attrs.empty())
		return;

	// parent
	std::string parent = m_wsdl.GetParent(NsPrefix() + ":" + Name());
	if (!parent.empty())
		fprintf(fil, "%s// parent: %s\n", prefix.c_str(), parent.c_str());

	// class
	fprintf(fil, "%sclass %s\n%s{\n", prefix.c_str(), CName().c_str(), prefix.c_str());

	// internal types
	try
	{
		const std::list<std::string>& ref = m_wsdl.ComplexTypes(NsPrefix());
		for (std::list<std::string>::const_iterator it = ref.begin(); it != ref.end(); it++)
		{
			std::string name = NsPrefix() + ":" + *it;
			std::string parent = m_wsdl.GetParent(name);
			if (parent == Name())
			{
				fprintf(fil, "%s// internal complexType: %s\n", prefix.c_str(), it -> c_str());
				fprintf(fil, "%spublic:\n", prefix.c_str());
				Xml::XmlNode n = m_wsdl.FindComplexType( name );
				ComplexType t(m_wsdl, n, Split(name).Type, Split(name).NS);
				t.CreateInterface(fil, lvl + 1);
			}
		}
	} catch (const Exception&) {}
	try
	{
		const std::list<std::string>& ref = m_wsdl.SimpleTypes(NsPrefix());
		for (std::list<std::string>::const_iterator it = ref.begin(); it != ref.end(); it++)
		{
			std::string name = NsPrefix() + ":" + *it;
			std::string parent = m_wsdl.GetParent(name);
			if (parent == Name())
			{
				fprintf(fil, "%s// internal simpleType: %s\n", prefix.c_str(), it -> c_str());
				fprintf(fil, "%spublic:\n", prefix.c_str());
				Xml::XmlNode n = m_wsdl.FindSimpleType( name );
				SimpleType t(m_wsdl, n, Split(name).Type, Split(name).NS);
				t.CreateInterface(fil, lvl + 1);
			}
		}
	} catch (const Exception&) {}

	// begin methods
	fprintf(fil, "%spublic:\n", prefix.c_str());
	// default constructor
	fprintf(fil, "%s\t%s();\n", prefix.c_str(), CName().c_str());
	// create from Xml::XmlNode constructor
	fprintf(fil, "%s\t%s(const Xml::XmlNode& it);\n", prefix.c_str(), CName().c_str());
	// destructor
	fprintf(fil, "%s\tvirtual ~%s() {}\n", prefix.c_str(), CName().c_str());
	fprintf(fil, "\n");
	// From Xml
	fprintf(fil, "%s\tvoid FromXml(const Xml::XmlNode& it);\n", prefix.c_str());
	fprintf(fil, "\n");
	// ToString
	fprintf(fil, "%s\tconst std::string ToString(const std::string& name) const;\n", prefix.c_str());
	fprintf(fil, "\n");
	// IsSet
	{
		fprintf(fil, "%s\tbool IsSet() const;\n", prefix.c_str());
		fprintf(fil, "\n");
	}
	// get/set methods
	for (std::list<std::pair<std::string, Xml::XmlNode> >::const_iterator it = members.begin(); it != members.end(); it++)
	{
		const std::pair<std::string, Xml::XmlNode>& ref = *it;
		Element el(m_wsdl, ref.second, "", NsPrefix());
		if (el.IsVector() || el.IsComplexType())
		{
			fprintf(fil, "%s\tconst %s Get%s() const;\n", prefix.c_str(), el.CType_out().c_str(), el.CName1up().c_str());
			fprintf(fil, "%s\t%s Get%s();\n", prefix.c_str(), el.CType_out().c_str(), el.CName1up().c_str());
			fprintf(fil, "%s\tvoid Set%s(%s %s);\n",
				prefix.c_str(), el.CName1up().c_str(), el.CType_in().c_str(), el.CName().c_str());
		}
		else
		{
			fprintf(fil, "%s\t%s Get%s() const;\n", prefix.c_str(), el.CType_out().c_str(), el.CName1up().c_str());
			fprintf(fil, "%s\tvoid Set%s(%s %s);\n", prefix.c_str(), el.CName1up().c_str(),
				el.CType_in().c_str(), el.CName().c_str());
		}
		fprintf(fil, "\n");
	}
	for (std::list<std::pair<std::string, Xml::XmlNode> >::const_iterator it = attrs.begin(); it != attrs.end(); it++)
	{
		const std::pair<std::string, Xml::XmlNode>& ref = *it;
		Attribute attr(m_wsdl, ref.second, ref.first, NsPrefix());

		fprintf(fil, "%s\t%s AttrGet%s() const;\n", prefix.c_str(), attr.CType_out().c_str(), attr.CName1up().c_str());
		fprintf(fil, "%s\tvoid AttrSet%s(%s %s);\n", prefix.c_str(), attr.CName1up().c_str(),
			attr.CType_in().c_str(), attr.CName().c_str());
	}
	if (!attrs.empty())
		fprintf(fil, "\n");
	// is_set methods
	for (std::list<std::pair<std::string, Xml::XmlNode> >::const_iterator it = members.begin(); it != members.end(); it++)
	{
		const std::pair<std::string, Xml::XmlNode>& ref = *it;
		Element el(m_wsdl, ref.second, "", NsPrefix());
		if (el.IsVector() || el.IsComplexType())
		{
			// _is_set is calculated
			fprintf(fil, "%s\tbool %sIsSet() const;\n", prefix.c_str(), el.CName1up().c_str(), el.CName().c_str());
		}
		else
		{
			fprintf(fil, "%s\tbool %sIsSet() const { return _element_%s_is_set; }\n", prefix.c_str(), el.CName1up().c_str(), el.CName().c_str());
		}
	}
	if (!members.empty())
		fprintf(fil, "\n");
	for (std::list<std::pair<std::string, Xml::XmlNode> >::const_iterator it = attrs.begin(); it != attrs.end(); it++)
	{
		const std::pair<std::string, Xml::XmlNode>& ref = *it;
		Attribute attr(m_wsdl, ref.second, ref.first, NsPrefix());

		fprintf(fil, "%s\tbool Attr%sIsSet() const { return _attribute_%s_is_set; }\n", prefix.c_str(), attr.CName1up().c_str(), attr.CName().c_str());
	}
	if (!attrs.empty())
		fprintf(fil, "\n");
	// private
	fprintf(fil, "%sprivate:\n", prefix.c_str());
	// elements
	for (std::list<std::pair<std::string, Xml::XmlNode> >::const_iterator it = members.begin(); it != members.end(); it++)
	{
		const std::pair<std::string, Xml::XmlNode>& ref = *it;
		Element el(m_wsdl, ref.second, "", NsPrefix());
		std::string doc = el.Documentation();
		if (!doc.empty())
			fprintf(fil, "%s\t/** %s */\n", prefix.c_str(), doc.c_str());
		fprintf(fil, "%s\t%s _element_%s;\n", prefix.c_str(), el.CType_decl().c_str(), el.CName().c_str());
	}
	for (std::list<std::pair<std::string, Xml::XmlNode> >::const_iterator it = attrs.begin(); it != attrs.end(); it++)
	{
		const std::pair<std::string, Xml::XmlNode>& ref = *it;
		Attribute attr(m_wsdl, ref.second, ref.first, NsPrefix());

		fprintf(fil, "%s\t%s _attribute_%s;\n", prefix.c_str(), attr.CType_decl().c_str(), attr.CName().c_str());
	}
	// is_set members
	for (std::list<std::pair<std::string, Xml::XmlNode> >::const_iterator it = members.begin(); it != members.end(); it++)
	{
		const std::pair<std::string, Xml::XmlNode>& ref = *it;
		Element el(m_wsdl, ref.second, "", NsPrefix());
		if (el.IsVector() || el.IsComplexType())
		{
			// _is_set is calculated
		}
		else
		{
			fprintf(fil, "%s\tbool _element_%s_is_set;\n", prefix.c_str(), el.CName().c_str());
		}
	}
	for (std::list<std::pair<std::string, Xml::XmlNode> >::const_iterator it = attrs.begin(); it != attrs.end(); it++)
	{
		const std::pair<std::string, Xml::XmlNode>& ref = *it;
		Attribute attr(m_wsdl, ref.second, ref.first, NsPrefix());

		fprintf(fil, "%s\tbool _attribute_%s_is_set;\n", prefix.c_str(), attr.CName().c_str());
	}
	//
	fprintf(fil, "%s};\n", prefix.c_str()); // End of class '%s'\n", CName().c_str());
	fprintf(fil, "\n");
}


void ComplexType::reset_members(std::list<std::pair<std::string, Xml::XmlNode> >& members, std::list<std::pair<std::string, Xml::XmlNode> >& attrs, FILE *fil) const
{
	std::string prefix = ": ";
	for (std::list<std::pair<std::string, Xml::XmlNode> >::const_iterator it = members.begin(); it != members.end(); it++)
	{
		const std::pair<std::string, Xml::XmlNode>& ref = *it;
		Element el(m_wsdl, ref.second, "", NsPrefix());
		std::string def = el.DefaultValue();
		if (el.IsBuiltin() && !def.empty())
		{
			fprintf(fil, "%s_element_%s(%s)\n", prefix.c_str(), el.CName().c_str(), el.DefaultValue().c_str());
			prefix = ", ";
		}
	}
	for (std::list<std::pair<std::string, Xml::XmlNode> >::const_iterator it = members.begin(); it != members.end(); it++)
	{
		const std::pair<std::string, Xml::XmlNode>& ref = *it;
		Element el(m_wsdl, ref.second, "", NsPrefix());
		if (el.IsVector() || el.IsComplexType())
		{
			// _is_set is calculated
		}
		else
		{
			fprintf(fil, "%s_element_%s_is_set(false)\n", prefix.c_str(), el.CName().c_str());
			prefix = ", ";
		}
	}
	for (std::list<std::pair<std::string, Xml::XmlNode> >::const_iterator it = attrs.begin(); it != attrs.end(); it++)
	{
		const std::pair<std::string, Xml::XmlNode>& ref = *it;
		Attribute attr(m_wsdl, ref.second, "", NsPrefix());
		std::string def = attr.DefaultValue();
		if (attr.IsBuiltin() && !def.empty())
		{
			fprintf(fil, "%s_attribute_%s(%s)\n", prefix.c_str(), attr.CName().c_str(), attr.DefaultValue().c_str());
			prefix = ", ";
		}
	}
	for (std::list<std::pair<std::string, Xml::XmlNode> >::const_iterator it = attrs.begin(); it != attrs.end(); it++)
	{
		const std::pair<std::string, Xml::XmlNode>& ref = *it;
		Attribute attr(m_wsdl, ref.second, "", NsPrefix());
		fprintf(fil, "%s_attribute_%s_is_set(false)\n", prefix.c_str(), attr.CName().c_str());
		prefix = ", ";
	}
}


void ComplexType::from_xml(std::list<std::pair<std::string, Xml::XmlNode> >& members, std::list<std::pair<std::string, Xml::XmlNode> >& attrs, FILE *fil) const
{
	for (std::list<std::pair<std::string, Xml::XmlNode> >::const_iterator it = members.begin(); it != members.end(); it++)
	{
		const std::pair<std::string, Xml::XmlNode>& ref = *it;
		Element el(m_wsdl, ref.second, "", NsPrefix());
		if (!el.IsOptional())
		{
			fprintf(fil, "\tif (!it.Exists(\"%s\"))\n", el.Name().c_str());
			fprintf(fil, "\t\tthrow Xml::XmlException(\"missing required node in element '%s': %s\");\n", m_name.c_str(), el.Name().c_str());
		}
		if (el.IsVector())
		{
			fprintf(fil, "\t{\n");
			fprintf(fil, "\t\tXml::XmlNode n(it, \"%s\");\n", el.Name().c_str());
			fprintf(fil, "\t\twhile (n)\n");
			fprintf(fil, "\t\t{\n");
			if (el.IsComplexType())
			{
				fprintf(fil, "\t\t\t%s tmp(n);\n", el.CType().c_str());
				fprintf(fil, "\t\t\t_element_%s.push_back( tmp );\n", el.CName().c_str());
			}
			else
			if (el.BaseType() == "std::string")
			{
				fprintf(fil, "\t\t\t_element_%s.push_back( n.GetContent() );\n", el.CName().c_str());
			}
			else
			{
				fprintf(fil, "\t\t\t_element_%s.push_back( atoi(n.GetContent().c_str()) );\n", el.CName().c_str());
			}
			fprintf(fil, "\t\t\t++n;\n");
			fprintf(fil, "\t\t}\n");
			fprintf(fil, "\t}\n");
		}
		else
		if (el.IsComplexType())
		{
			fprintf(fil, "\t{\n");
			fprintf(fil, "\t\tXml::XmlNode n(it, \"%s\");\n", el.Name().c_str());
			fprintf(fil, "\t\tif (n)\n");
			fprintf(fil, "\t\t\t_element_%s.FromXml( n );\n", el.CName().c_str());
			fprintf(fil, "\t}\n");
		}
		else
		if (el.BaseType() == "std::string")
		{
			if (el.IsOptional())
				fprintf(fil, "\tif (it.Exists(\"%s\"))\n\t", el.Name().c_str());
			fprintf(fil, "\tSet%s( it[\"%s\"] );\n", el.CName1up().c_str(), el.Name().c_str());
		}
		else
		{
			if (el.IsOptional())
				fprintf(fil, "\tif (it.Exists(\"%s\"))\n\t", el.Name().c_str());
			fprintf(fil, "\tSet%s( atoi(std::string(it[\"%s\"]).c_str()) );\n", el.CName1up().c_str(), el.Name().c_str());
		}
	}
	for (std::list<std::pair<std::string, Xml::XmlNode> >::const_iterator it = attrs.begin(); it != attrs.end(); it++)
	{
		const std::pair<std::string, Xml::XmlNode>& ref = *it;
		Attribute attr(m_wsdl, ref.second, "", NsPrefix());

		if (attr.IsOptional())
		{
			fprintf(fil, "\tif (it.PropertyExists(\"%s\"))\n\t", attr.Name().c_str());
		}
		if (attr.BaseType() == "std::string")
			fprintf(fil, "\tAttrSet%s(it.GetProperty(\"%s\"));\n", attr.CName1up().c_str(), attr.Name().c_str());
		else
			fprintf(fil, "\tAttrSet%s( atoi(it.GetProperty(\"%s\").c_str()) );\n", attr.CName1up().c_str(), attr.Name().c_str());
	}
}


void ComplexType::to_string(std::list<std::pair<std::string, Xml::XmlNode> >& members, std::list<std::pair<std::string, Xml::XmlNode> >& attrs, FILE *fil) const
{
	if (attrs.empty())
	{
		fprintf(fil, "\tstd::string r = \"<\" + name + \">\";\n");
	}
	else
	{
		fprintf(fil, "\tstd::string r = \"<\" + name;\n");

		for (std::list<std::pair<std::string, Xml::XmlNode> >::const_iterator it = attrs.begin(); it != attrs.end(); it++)
		{
			const std::pair<std::string, Xml::XmlNode>& ref = *it;
			Attribute attr(m_wsdl, ref.second, "", NsPrefix());

			if (attr.IsOptional())
			{
				fprintf(fil, "\tif (Attr%sIsSet())\n\t", attr.CName1up().c_str());
			}
			if (attr.BaseType() == "std::string")
				fprintf(fil, "\tr += \" %s='\" + AttrGet%s() + \"'\";\n", attr.Name().c_str(), attr.CName1up().c_str());
			else
			{
				fprintf(fil, "\t{\n");
				fprintf(fil, "\t\tchar slask[100];\n");
				fprintf(fil, "\t\tsprintf(slask, \"%%d\", AttrGet%s());\n", attr.CName1up().c_str());
				fprintf(fil, "\t\tr = r + \" %s='\" + slask + \"'\";\n", attr.Name().c_str());
				fprintf(fil, "\t}\n");
			}
		}
		fprintf(fil, "\tr += \">\";\n");
	}

	for (std::list<std::pair<std::string, Xml::XmlNode> >::const_iterator it = members.begin(); it != members.end(); it++)
	{
		const std::pair<std::string, Xml::XmlNode>& ref = *it;
		Element el(m_wsdl, ref.second, "", NsPrefix());
		if (el.IsVector())
		{
			fprintf(fil, "\tfor (%s<%s>::const_iterator it = _element_%s.begin(); it != _element_%s.end(); it++)\n",
				m_wsdl.GetApplication().ListType().c_str(),
				el.CType().c_str(), el.CName().c_str(), el.CName().c_str());
			fprintf(fil, "\t{\n");

			if (el.IsComplexType())
			{
				fprintf(fil, "\t\tconst %s& ref = *it;\n", el.CType().c_str());
				fprintf(fil, "\t\tr += ref.ToString(\"%s\");\n", el.Name().c_str());
			}
			else
			if (el.BaseType() == "std::string")
			{
				fprintf(fil, "\t\tr += \"<%s>\" + *it + \"</%s>\";\n",
					el.Name().c_str(), el.Name().c_str());
			}
			else
			{
				fprintf(fil, "\t\t{\n");
				fprintf(fil, "\t\t\tchar slask[100];\n");
				fprintf(fil, "\t\t\tsprintf(slask, \"%%d\", *it);\n");
				fprintf(fil, "\t\t\tr = r + \"<%s>\" + slask + \"</%s>\";\n",
					el.Name().c_str(), el.Name().c_str());
				fprintf(fil, "\t\t}\n");
			}

			fprintf(fil, "\t}\n");
		}
		else
		if (el.IsComplexType())
		{
			fprintf(fil, "\tr += _element_%s.ToString(\"%s\");\n", el.CName().c_str(), el.Name().c_str());
		}
		else
		if (el.BaseType() == "std::string")
		{
			if (el.IsOptional())
				fprintf(fil, "\tif (_element_%s_is_set)\n\t", el.CName().c_str());
			fprintf(fil, "\tr += \"<%s>\" + _element_%s + \"</%s>\";\n",
				el.Name().c_str(), el.CName().c_str(), el.Name().c_str());
		}
		else
		{
			if (el.IsOptional())
				fprintf(fil, "\tif (_element_%s_is_set)\n", el.CName().c_str());
			fprintf(fil, "\t{\n");
			fprintf(fil, "\t\tchar slask[100];\n");
			fprintf(fil, "\t\tsprintf(slask, \"%%d\", _element_%s);\n", el.CName().c_str());
			fprintf(fil, "\t\tr = r + \"<%s>\" + slask + \"</%s>\";\n",
				el.Name().c_str(), el.Name().c_str());
			fprintf(fil, "\t}\n");
		}
	}
	fprintf(fil, "\tr += \"</\" + name + \">\";\n");
	fprintf(fil, "\treturn r;\n");
}


void ComplexType::CreateImplementation(FILE *fil) const
{
	std::list<std::pair<std::string, Xml::XmlNode> > members = GetMembers();
	std::list<std::pair<std::string, Xml::XmlNode> > attrs = GetAttributes();
	if (members.empty() && attrs.empty())
		return;

	std::string parent = m_wsdl.GetParent(NsPrefix() + ":" + Name());
	std::string tmp;
	while (!parent.empty())
	{
		tmp = Split(parent).CName1up + "::" + tmp;
		parent = m_wsdl.GetParent(NsPrefix() + ":" + parent);
	}
	std::string classname = tmp + CName();

	// default constructor
	fprintf(fil, "// -----------------------------------------------------------------------------\n");
	fprintf(fil, "%s::%s()\n", classname.c_str(), CName().c_str());
	reset_members(members, attrs, fil);
	fprintf(fil, "{\n");
	fprintf(fil, "}\n\n\n");
	// Xml::XmlNode constructor
	fprintf(fil, "// -----------------------------------------------------------------------------\n");
	fprintf(fil, "%s::%s(const Xml::XmlNode& it)\n", classname.c_str(), CName().c_str());
	reset_members(members, attrs, fil);
	fprintf(fil, "{\n");
	fprintf(fil, "\tFromXml(it);\n");
	fprintf(fil, "}\n\n\n");
	// From Xml
	fprintf(fil, "// -----------------------------------------------------------------------------\n");
	fprintf(fil, "void %s::FromXml(const Xml::XmlNode& it)\n{\n", classname.c_str());
	from_xml(members, attrs, fil);
	fprintf(fil, "}\n\n\n");
	// ToString
	fprintf(fil, "// -----------------------------------------------------------------------------\n");
	fprintf(fil, "const std::string %s::ToString(const std::string& name) const\n{\n", classname.c_str());
	to_string(members, attrs, fil);
	fprintf(fil, "}\n\n\n");
	// IsSet
	{
		fprintf(fil, "// -----------------------------------------------------------------------------\n");
		fprintf(fil, "bool %s::IsSet() const\n{\n", classname.c_str());
		std::string prefix = "\treturn ";
		for (std::list<std::pair<std::string, Xml::XmlNode> >::const_iterator it = members.begin(); it != members.end(); it++)
		{
			const std::pair<std::string, Xml::XmlNode>& ref = *it;
			Element el(m_wsdl, ref.second, "", NsPrefix());
			fprintf(fil, "%s%sIsSet()", prefix.c_str(), el.CName1up().c_str());
			prefix = " || ";
		}
		for (std::list<std::pair<std::string, Xml::XmlNode> >::const_iterator it = attrs.begin(); it != attrs.end(); it++)
		{
			const std::pair<std::string, Xml::XmlNode>& ref = *it;
			Attribute attr(m_wsdl, ref.second, "", NsPrefix());
			fprintf(fil, "%sAttr%sIsSet()", prefix.c_str(), attr.CName1up().c_str());
			prefix = " || ";
		}
		fprintf(fil, ";\n");
		fprintf(fil, "}\n\n\n");
	}
	// get/set
	for (std::list<std::pair<std::string, Xml::XmlNode> >::const_iterator it = members.begin(); it != members.end(); it++)
	{
		const std::pair<std::string, Xml::XmlNode>& ref = *it;
		Element el(m_wsdl, ref.second, "", NsPrefix());
		fprintf(fil, "// -----------------------------------------------------------------------------\n");
		if (el.IsVector() || el.IsComplexType())
		{
			fprintf(fil, "const %s %s::Get%s() const\n{\n", el.CType_out().c_str(), classname.c_str(), el.CName1up().c_str());
			fprintf(fil, "\treturn _element_%s;\n", el.CName().c_str());
			fprintf(fil, "}\n\n\n");
			fprintf(fil, "%s %s::Get%s()\n{\n", el.CType_out().c_str(), classname.c_str(), el.CName1up().c_str());
			fprintf(fil, "\treturn _element_%s;\n", el.CName().c_str());
			fprintf(fil, "}\n\n\n");
			fprintf(fil, "void %s::Set%s(%s %s)\n{\n\t_element_%s = %s;\n}\n\n\n",
				classname.c_str(),
				el.CName1up().c_str(), el.CType_in().c_str(), el.CName().c_str(),
				el.CName().c_str(), el.CName().c_str());
		}
		else
		{
			fprintf(fil, "%s %s::Get%s() const\n{\n", el.CType_out().c_str(), classname.c_str(), el.CName1up().c_str());
			fprintf(fil, "\treturn _element_%s;\n", el.CName().c_str());
			fprintf(fil, "}\n\n\n");
			fprintf(fil, "void %s::Set%s(%s %s)\n{\n", classname.c_str(), el.CName1up().c_str(),
				el.CType_in().c_str(), el.CName().c_str());
			fprintf(fil, "\t_element_%s = %s;\n", el.CName().c_str(), el.CName().c_str());
			fprintf(fil, "\t_element_%s_is_set = true;\n");
			fprintf(fil, "}\n\n\n");
		}
	}
	// is_set
	for (std::list<std::pair<std::string, Xml::XmlNode> >::const_iterator it = members.begin(); it != members.end(); it++)
	{
		const std::pair<std::string, Xml::XmlNode>& ref = *it;
		Element el(m_wsdl, ref.second, "", NsPrefix());
		if (el.IsVector() || el.IsComplexType())
		{
			// _is_set is calculated
			fprintf(fil, "// -----------------------------------------------------------------------------\n");
			fprintf(fil, "bool %s::%sIsSet() const\n{\n", classname.c_str(), el.CName1up().c_str());
			if (el.IsVector())
				fprintf(fil, "\treturn !_element_%s.empty();\n", el.CName().c_str());
			else
				fprintf(fil, "\treturn _element_%s.IsSet();\n", el.CName().c_str());
			fprintf(fil, "}\n\n\n");
		}
		else
		{
		}
	}
	// attribute get/set
	for (std::list<std::pair<std::string, Xml::XmlNode> >::const_iterator it = attrs.begin(); it != attrs.end(); it++)
	{
		const std::pair<std::string, Xml::XmlNode>& ref = *it;
		Attribute attr(m_wsdl, ref.second, "", NsPrefix());
		fprintf(fil, "// -----------------------------------------------------------------------------\n");
		{
			fprintf(fil, "%s %s::AttrGet%s() const\n{\n", attr.CType_out().c_str(), classname.c_str(), attr.CName1up().c_str());
			fprintf(fil, "\treturn _attribute_%s;\n", attr.CName().c_str());
			fprintf(fil, "}\n\n\n");
			fprintf(fil, "void %s::AttrSet%s(%s %s)\n{\n", classname.c_str(), attr.CName1up().c_str(),
				attr.CType_in().c_str(), attr.CName().c_str());
			fprintf(fil, "\t_attribute_%s = %s;\n", attr.CName().c_str(), attr.CName().c_str());
			fprintf(fil, "\t_attribute_%s_is_set = true;\n");
			fprintf(fil, "}\n\n\n");
		}
	}
}


